
//
//        Copyright 2010 Johan Dahlberg. All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions
//  are met:
//
//    1. Redistributions of source code must retain the above copyright notice,
//       this list of conditions and the following disclaimer.
//
//    2. Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//
//  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
//  AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
//  THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
//  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

var inherits              = require("util").inherits;

var EventEmitter          = require("events").EventEmitter;

var kernel                = require("./kernel");
var Dispatcher            = require("./dispatcher").Dispatcher;
var Component             = require("./component").Component;



function Actor(params) {
  EventEmitter.call(this);
  Dispatcher.call(this);
  Component.call(this);

  params = params || {};

  for (var param in params) {
    if (this.hasParam(param) == false) {
      throw new Error("invalid param `" + param + "`");
    }
  }

  kernel.registerLocal(this, params);

  this.monitor = require("./monitor").createMonitor(this._ref.id.toString());
}

exports.Actor = Actor;
inherits(Actor, EventEmitter);



(function() {
  function cp(s, t) { for (var k in s) { t[k] = s[k]; } }
  cp(Dispatcher.prototype, Actor.prototype);
  cp(Component.prototype, Actor.prototype);
})();



Actor.findCallback = Dispatcher.findCallback;
Actor.defineHandler = Dispatcher.defineHandler;
Actor.isHandler = Dispatcher.isHandler;
Actor.defineParam = Component.defineParam;
Actor.isParam = Component.isParam;



Actor.define = function(ctor) {
  ctor = ctor || function() { Actor.call(this); };
  inherits(ctor, Actor);
  return ctor;
};


Actor.isActorConstructor = function(ctor) {

  function isactorctor(fn) {
    if (typeof fn !== "function") return false;
    if (fn === Actor) return true;
    if (typeof fn.super_ !== "function") return false;
    return isactorctor(fn.super_);
  }

  return isactorctor(ctor);
};



Object.defineProperty(Actor.prototype, "sync", {
  set: function(fn) { Actor.defineHandler(this, fn.name, { sync: fn }); }
});



Object.defineProperty(Actor.prototype, "async", {
  set: function(fn) { Actor.defineHandler(this, fn.name, { async: fn }); }
});



Object.defineProperty(Actor.prototype, "notif", {
  set: function(fn) { Actor.defineHandler(this, fn.name, { notif: fn }); }
});



Actor.defineParam(Actor.prototype, "ref", {
  desc: "Gets the REF for actor", readonly: true,
  get: function() { return this._ref }
});



Actor.defineHandler(Actor.prototype, "desc", {
  desc: "Gets a descrption for specified `dispatcher` or `param`",
  async: function(name, callback) {
    var callback = Actor.findCallback(arguments);

    if (typeof name !== "string") {
      return callback(new Error("bad argument, `name`, expected String"));
    }


    callback(null, Actor.isHandler(this[name]));
  }
});



Actor.defineHandler(Actor.prototype, "dispatching", {
  async: function(name, callback) {
    if (typeof name !== "string") {
      return callback(new Error("bad argument, `name`, expected String"));
    }
    callback(null, Actor.isHandler(this[name]));
  }
});



Actor.defineHandler(Actor.prototype, "get", {
  desc: "Get the value of param `param`",
  sync: function(param) {
    var callback = Actor.findCallback(arguments);
    var prop;

    if (typeof param !== "string") {
      return callback(new Error("bad argument, `param`, expected String"));
    }

    if (this.hasParam(param) == false) {
      return callback(new Error("param `" + param + "` was not found"));
    }

    callback(null, this.__get__(param));
  }
});



Actor.defineHandler(Actor.prototype, "set", {
  desc: "Set `value` of param `param`",
  sync: function(param, value) {
    var callback = Actor.findCallback(arguments);
    var prop;

    if (typeof param !== "string") {
      return callback(new Error("bad argument, `param`, expected String"));
    }

    prop = "__set_" + param + "__";

    if (typeof this[prop] != "function") {
      return callback(new Error("param `" + param + "` was not found"));
    }

    this[prop](value, callback);
  }
});



Actor.defineHandler(Actor.prototype, "filter", {
  desc: "Gets all available params based on provided `filter`",
  async: function(filter) {
    var callback = Actor.findCallback(arguments);
    var result;

    result = this.__filter__(filter);

    callback && callback(null, result);
  }
});



Actor.defineHandler(Actor.prototype, "params", {
  desc: "Set all params in object `params`",
  sync: function(params) {
    var callback = Actor.findCallback(arguments);
    var self = this;
    var keys;
    var index;

    if (typeof params !== "object") {
      return callback("bad argument, `params`, expected Object");
    }

    function validate(param, result, next) {
      self.__validate__(param, params[param],
        function(err, value) {
          if (err) return next(err);
          result[param] = value;
          next(null, result);
      });
    }

    keys = Object.keys(params);
    index = keys.length;

    (function validatenext(err, result) {
      var param = keys[--index];
      if (err) { return callback(err); }
      if (param) {
        validate(param, result, validatenext);
      } else {
        index = keys.length;
        while (index--) {
          param = keys[index];
          self.__set__(param, result[param]);
        }
        callback(null, result);
      }
    })(null, {});
  }
});



Actor.defineHandler(Actor.prototype, "exit", {
  sync: function(reason, callback) {
    this.destroy(reason || "NORMAL");
    process.nextTick(callback);
  }
});



Actor.defineHandler(Actor.prototype, "ping", {
  sync: function(callback) {
    callback();
  }
});



Actor.defineHandler(Actor.prototype, "kill", {
  notif: function(code) {
    this.destroy(code);
  }
});



Actor.defineHandler(Actor.prototype, "log", {
  notif: function() {
    this.monitor.log.apply(this.monitor, arguments);
  }
});



Actor.defineHandler(Actor.prototype, "verb", {
  notif: function() {
    this.monitor.verbose.apply(this.monitor, arguments);
  }
});



Actor.defineHandler(Actor.prototype, "err", {
  notif: function() {
    this.monitor.error.apply(this.monitor, arguments);
  }
});



Actor.prototype.isFirstResponder = function() {
  return kernel.isFirstResponder(this.ref);
};



Actor.prototype.self = function() {
  return this._ref;
};



Actor.prototype.destroy = function(err) {
  this._inbox = null;
  
  Dispatcher.destroy(this);
  Component.destroy(this);

  this.ondestroy && this.ondestroy(err);

  kernel.unregisterLocal(this, err);
};

