
//
//        Copyright 2010 Johan Dahlberg. All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions
//  are met:
//
//    1. Redistributions of source code must retain the above copyright notice,
//       this list of conditions and the following disclaimer.
//
//    2. Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//
//  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
//  AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
//  THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
//  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

module.exports = exports = new (require("events")).EventEmitter;


function Component() {
  this.__mappers__ = null;
}

exports.Component = Component;


Component.defineParam = function(target, param, descriptor) {
  var validator;
  var refval;
  var refget;
  var refset;

  if (typeof target !== "object") {
    throw new Error("bad argument, `target`, expected Object");
  }

  if (typeof param !== "string") {
    throw new Error("bad argument, `param`, expected String");
  }

  refval = "__val_" + param + "__";
  refget = "__get_" + param + "__";
  refset = "__set_" + param + "__";

  if (target.hasOwnProperty(param) ||
      target.hasOwnProperty(refval) ||
      target.hasOwnProperty(refget) ||
      target.hasOwnProperty(refset)) {
    throw new Error("bad argument, `param`, already defined");
  }

  if (typeof descriptor !== "object") {
    throw new Error("bad argument, `descriptor`, expected Object");
  }

  if (descriptor.validator && descriptor.type) {
    throw new Error("cannot contain both validator and type");
  }


  validator = require("./param").getValidator(descriptor);

  function getter() {
    return this[refget];
  }

  function setter(value, callback) {
    var self = this;
    validator.call(this, value, function(err, value) {
      if (err) return callback && callback(err);
      if (self.__set__) self.__set__(param, value);
      return callback && callback(err, value);
    });
  }

  Object.defineProperty(target, param, {
    enumerable: true,
    configurable: false,
    get: descriptor.get || getter,
    set: setter
  });

  target[refget] = descriptor.get || descriptor.initial || null;
  target[refset] = setter;
  target[refval] = validator;
};



Component.isParam = function(target, param) {
  return target ? typeof target["__set_" + param + "__"] == "function" : false;
};



Component.destroy = function(target) {
  target.__mappers__ = null;
};



Component.prototype.hasParam = function(param) {
  return Component.isParam(this, param);
};



Component.prototype.map = function(param, dest, prop) {

  if (this.hasParam(param) == false) {
    throw new Error("bad argument, `param`, not found");
  }

  if (!dest) {
    throw new Error("bad argument, expected `dest`");
  }

  if (typeof prop == "string" || typeof prop == "function") {
    this.__attachobjmapper__(param, dest, prop);
  } else if (typeof dest == "function"){
    this.__attachmapper__(param, dest);
  } else {
    throw new Error("bad arguments");
  }
};



Component.prototype.unmap = function(param, dest, prop) {

  if (this.hasParam(param) == false) {
    throw new Error("bad argument, `param`, not found");
  }

  if (!dest) {
    throw new Error("bad argument, expected `dest`");
  }

  if (typeof prop == "string" || typeof prop == "function") {
    this.__deattachobjmapper__(param, dest, prop);
  } else if (typeof dest == "function") {
    this.__deattachmapper__(param, dest);
  } else {
    throw new Error("bad arguments");
  }
};



Component.prototype.__filter__ = function(filter) {
  var result = {};
  var param;
  var re;
  var m;
  
  if (typeof filter == "string" && filter !== "*") {
    re = new RegExp("^" + filter);
    for (var prop in this) {
      if (m = (/__get_([A-Za-z0-9\$\_]+)__$/).exec(prop)) {
        console.log(m);
        console.log(re);
        console.log(re.test((param = m[1])));
        if (re.test((param = m[1]))) {
          console.log(m);
          result[param] = this.__get__(param);
        }
      }
    }
  } else {
    for (var prop in this) {
      if (m = (/__get_([A-Za-z0-9\$\_]+)__$/).exec(prop)) {
        result[(param = m[1])] = this.__get__(param);
      }
    }
  }

  return result;
};



Component.prototype.__attachmapper__ = function(param, callback) {
  var mappers = this.__mappers__;
  var initial = this.__get__(param);

  if (!mappers) {
    this.__mappers__ = mappers = [];
  }

  if (typeof mappers[param] == "undefined") {
    mappers[param] = [];
  }

  mappers[param].push(callback);
  
  callback(initial, initial);
};



Component.prototype.__deattachmapper__ = function(param, callback) {
  var mappers = this.__mappers__;
  var callbacks;
  var index;

  if (!mappers || !(callbacks = mappers[param])) {
    return;
  }

  if ((index = callbacks.indexOf(callback) == -1)) {
    return;
  }

  callbacks.splice(index, 1);

  if (callbacks.length == 0) {
    delete this.__mappers__[param];
  }
};



Component.prototype.__attachobjmapper__ = function(param, obj, prop) {
  var listener;

  if (typeof prop == "function") {
    listener = function(value, old) { prop.call(obj, value, old); }
  } else {
    listener = function(value) { obj[prop] = value; }
  }

  listener._obj = obj;
  listener._prop = prop;

  this.__attachmapper__(param, listener);
};



Component.prototype.__deattachobjmapper__ = function(param, obj, prop) {
  var mappers = this.__mappers__;
  var callbacks;
  var callback;
  var index;

  if (!mappers || !(callbacks = mappers[param])) {
    return;
  }

  index = callbacks.length;

  if (prop) {
    while (index--) {
      callback = callbacks[index];
      if (callback._obj == obj && callback._prop == prop) {
        return this.__deattachmapper__(param, callback);
      }
    }
  } else {
    while (index--) {
      callback = callbacks[index];
      if (callback._obj == obj) {
        return this.__deattachmapper__(param, callback);
      }
    }
  }
};



Component.prototype.__set__ = function(param, value) {
  var refget = "__get_" + param + "__";
  var old = this[refget];
  var callbacks;
  var emitkey;

  value = value === void(0) ? null : value;
  this[refget] = value;

  if (this.__mappers__ && (callbacks = this.__mappers__[param])) {
    index = callbacks.length;
    while (index--) { 
      callbacks[index](value, old);
    }
  }
  emitkey = "param." + this._ref.id + "." + param;
  if (exports._events && exports._events[emitkey]) {
    exports.emit(emitkey, value, old);
  }
  if (this._name) {
    emitkey = "param." + this._name + "." + param;
    if (exports._events && exports._events[emitkey]) {
      exports.emit(emitkey, value, old);
    }
  }
};



Component.prototype.__get__ = function(param) {
  var refget = "__get_" + param + "__";
  if (typeof this[refget] == "function") {
    return this[refget]();
  } else {
    return this[refget];
  }
};



Component.prototype.__validate__ = function(param, value, callback) {
  var validator = this["__val_" + param + "__"];
  return validator.call(this, value, callback);
};