
//
//        Copyright 2010 Johan Dahlberg. All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions
//  are met:
//
//    1. Redistributions of source code must retain the above copyright notice,
//       this list of conditions and the following disclaimer.
//
//    2. Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//
//  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
//  AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
//  THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
//  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

var inherits              = require("util").inherits;

var EventEmitter          = require("events").EventEmitter;


function Dispatcher() {
  EventEmitter.call(this);

  this.__inbox__ = null;
  this.__mutexcount__ = 0;
  this.__mutexlock__ = null;
}

inherits(Dispatcher, EventEmitter);
exports.Dispatcher = Dispatcher;



Dispatcher.findCallback = function(args) {
  var index = args.length;
  while (index--) {
    if (typeof args[index] == "function") {
      return args[index];
    }
  }
};


Dispatcher.destroy = function(target) {
  target.__inbox__ = null;
  target.__mutexcount__ = 0;
  target.__mutexlock__ = null;
};


Dispatcher.isHandler = function(h) {
  return typeof h == "function" && h.name == "__handler__" || false;
};



Dispatcher.defineHandler = function(target, name, descriptor) {

  if (typeof target !== "object") {
    throw new Error("bad argument, `target`, expected Object");
  }

  if (target.hasOwnProperty(name)) {
    throw new Error("bad argument, `handler`, already defined");
  }

  if (typeof descriptor !== "object") {
    throw new Error("bad argument, `descriptor`, expected Object");
  }

  if (typeof descriptor.async == "function") {
    target[name] = defineAsyncHandler(descriptor.async);
  } else if (typeof descriptor.sync == "function") {
    target[name] = defineSyncHandler(descriptor.sync);
  } else if (typeof descriptor.notif == "function") {
    target[name] = defineNotifHandler(descriptor.notif);
  } else {
    throw new Error("bad argument, `descriptor`, expected async/sync or notif");
  }
};



Dispatcher.prototype.addHandler = function(name, type, handler) {
  var descriptor = {};

  if (typeof name !== "string") {
    throw new Error("bad argument, `name`, expected string");
  }

  if (this.hasHandler(name)) {
    throw new Error("handler `" + name + "` is already defined");
  }

  if (/^(sync|async|notif)$/i.test(type) == false) {
    throw new Error("bad argument, `type`, expected sync, async or notif");
  }

  if (typeof handler !== "function") {
    throw new Error("bad argument, `handler`, expected function");
  }

  descriptor[type] = handler;

  Actor.defineHandler(this, name, descriptor);
};



Dispatcher.prototype.dispatch = function(handler, args, opts, callback) {
  var disp;

  if (typeof opts == "function") {
    callback = opts;
    opts = null;
  }

  if (this.hasHandler(handler) == false) {
    callback(new Error("bad dispatcher `" + handler + "`"));
    return false;
  }

  if (Array.isArray(args) == false) {
    throw new Error("bad argument, `args`, expected Array");
  }

  opts = opts || {};

  if (callback) {
    args = args.slice(0);
    args[args.length] = callback;
  }

  switch (args.length) {
    case 0: return this[handler]();
    case 1: return this[handler](args[0]);
    case 2: return this[handler](args[0], args[1]);
    case 3: return this[handler](args[0], args[1], args[2]);
    case 4: return this[handler](args[0], args[1], args[2], args[3]);
    case 5: return this[handler](args[0], args[1], args[2], args[3], args[4]);
    default: return this[handler].apply(this, args);
  }
};



Dispatcher.prototype.hasHandler = function(name) {
  return Dispatcher.isHandler(this[name]);
};



Dispatcher.prototype.pause = function() {
  if (!this.__inbox__) {
    this.__inbox__ = [];
    this.emit("pause");
  }
};



Dispatcher.prototype.resume = function() {
  var inbox;
  var item;

  if ((inbox = this.__inbox__)) {
    this.__inbox__ = null;

    for (var i = 0, l = inbox.length; i < l; i++) {
      item = inbox[i];
      this._dispatchMessage(item.cmd, item.args, item.callback);
    }

    this.emit("resume");
  }
};



Dispatcher.prototype._lock = function(callback) {
  if (this.__mutexlock__) {
    if (this.__mutexlock__ instanceof Array) {
      this.__mutexlock__.push(callback);
    } else {
      this.__mutexlock__ = [callback, this._mutexlock];
    }
  } else if (!this.__inbox__) {
    this.__inbox__ = [];
    process.nextTick(callback);
  } else {
    this.__mutexlock__ = callback;
  }
};



Dispatcher.prototype._unlock = function() {
  var inbox;
  var next;

  if (this.__mutexlock__) {
    if (this.__mutexlock__ instanceof Array) {
      next = this.__mutexlock__.pop();
      if (!this.__mutexlock__.length) {
        this.__mutexlock__ = null;
      }
    } else {
      next = this.__mutexlock__;
      this.__mutexlock__ = null;
    }
    return next();
  }

  if (this.__inbox__) {
    inbox = this.__inbox__;
    this.__inbox__ = null;

    for (var i = 0, l = inbox.length; i < l; i++) {
      inbox[i]();
    }
  }
};



Dispatcher.prototype._incr = function() {
  this.__mutexcount__++;
};



Dispatcher.prototype._decr = function() {
  var next;

  this.__mutexcount__--;

  if (this.__mutexcount__ < 0) {
    this.emit("error", "mutex ref error");
    return;
  }

  if (this.__mutexcount__ == 0) {
    if (this.__mutexlock__) {
      if (this.__mutexlock__ instanceof Array) {
        next = this.__mutexlock__.pop();
        if (!this.__mutexlock__.length) {
          this.__mutexlock__ = null;
        }
      } else {
        next = this.__mutexlock__;
        this.__mutexlock__ = null;
      }
      return next();
    }
  }
};



function defineAsyncHandler(handler) {
  return function __handler__(a, b, c, d, e) {
    var args = arguments;
    var argslen = args.length;
    var self = this;
    var callback;
    var r;

    if (this.__inbox__) {
      this.__inbox__.push(function() {
        switch (argslen) {
          case 0: __handler__.call(self); break;
          case 1: __handler__.call(self, a); break;
          case 2: __handler__.call(self, a, b); break;
          case 3: __handler__.call(self, a, b, c); break;
          case 4: __handler__.call(self, a, b, c, d); break;
          case 5: __handler__.call(self, a, b, c, d, e); break;
          default: __handler__.apply(self, args);
        }
      });
      return false;
    }

    if (argslen && typeof args[argslen - 1] == "function") {
      callback = args[--argslen];
    }

    this._incr();

    r = function(a, b, c, d, e) {
      r = null;
      self._decr();
      if (callback) {
        switch (arguments.length) {
          case 0: callback(); break;
          case 1: callback(a); break;
          case 2: callback(a, b); break;
          case 3: callback(a, b, c); break;
          case 4: callback(a, b, c, d); break;
          case 5: callback(a, b, c, d, e); break;
          default: callback.apply(null, arguments); break;
        }
      }
    };

    // try {
      switch (argslen) {
        case 0: handler.call(self, r); break;
        case 1: handler.call(self, a, r); break;
        case 2: handler.call(self, a, b, r); break;
        case 3: handler.call(self, a, b, c, r); break;
        case 4: handler.call(self, a, b, c, d, r); break;
        case 5: handler.call(self, a, b, c, d, e, r); break;
        default: handler.apply(self, args.concat([r])); break;
      }
    // } catch (disperror) {
    //   console.log(disperror.stack)
    //   r && r(disperror);
    // }

    return true;
  };
}



function defineSyncHandler(handler) {
  return function __handler__(a, b, c, d, e) {
    var args = arguments;
    var argslen = args.length;
    var self = this;
    var callback;
    var r;

    if (this.__inbox__) {
      this.__inbox__.push(function() {
        switch (argslen) {
          case 0: __handler__.call(self); break;
          case 1: __handler__.call(self, a); break;
          case 2: __handler__.call(self, a, b); break;
          case 3: __handler__.call(self, a, b, c); break;
          case 4: __handler__.call(self, a, b, c, d); break;
          case 5: __handler__.call(self, a, b, c, d, e); break;
          default: __handler__.apply(self, args);
        }
      });
      return false;
    }

    if (argslen && typeof args[argslen - 1] == "function") {
      callback = args[--argslen];
    }
    
    r = function(a, b, c, d, e) {
      r = null;
      self._unlock();
      if (callback) {
        switch (arguments.length) {
          case 0: callback(); break;
          case 1: callback(a); break;
          case 2: callback(a, b); break;
          case 3: callback(a, b, c); break;
          case 4: callback(a, b, c, d); break;
          case 5: callback(a, b, c, d, e); break;
          default: callback.apply(null, arguments); break;
        }
      }
    };

    this._lock(function() {
      // try {
        switch (argslen) {
          case 0: handler.call(self, r); break;
          case 1: handler.call(self, a, r); break;
          case 2: handler.call(self, a, b, r); break;
          case 3: handler.call(self, a, b, c, r); break;
          case 4: handler.call(self, a, b, c, d, r); break;
          case 5: handler.call(self, a, b, c, d, e, r); break;
          default: handler.apply(self, args.concat([r])); break;
        }
      // } catch (disperror) {
      //   console.log(disperror.stack);
      //   r && r(disperror);
      // }
    });

    return true;
  };
}



function defineNotifHandler(handler) {
  return function __handler__(a, b, c, d, e) {
    var args = arguments;
    var argslen = args.length;
    var self = this;
    var callback;

    if (argslen && typeof args[argslen - 1] == "function") {
      callback = args[--argslen];
    }

    // try {
      switch (argslen) {
        case 0: handler.call(self); break;
        case 1: handler.call(self, a); break;
        case 2: handler.call(self, a, b); break;
        case 3: handler.call(self, a, b, c); break;
        case 4: handler.call(self, a, b, c, d); break;
        case 5: handler.call(self, a, b, c, d, e); break;
        default: handler.apply(self, args.concat([r])); break;
      }
    // } catch (disperror) {
    //   console.log(disperror.stack);
    //   callback && callback(disperror);
    //   return false;
    // }

    if (callback) {
      callback();
    }

    return true;
  };
}