
//
//        Copyright 2010 Johan Dahlberg. All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions
//  are met:
//
//    1. Redistributions of source code must retain the above copyright notice,
//       this list of conditions and the following disclaimer.
//
//    2. Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//
//  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
//  AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
//  THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
//  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

var inherits              = require("util").inherits;

var EventEmitter          = require("events").EventEmitter;

var Dispatcher            = require("./dispatcher").Dispatcher;
var Component             = require("./component").Component;


var ID_POOL_MAX = 0xFFFF;
var ID_POOL_INITIAL_SIZE = 100;
var ID_POOL_INCREASE_STEP = 1.2;

var Kernel;
var locals = {};
var remotes = {};
var names = {};
var inboxes = {};
var namedinboxes = {};
var protocolWaitlist = {};
var protocolHandlers = {};


module.exports = exports = new EventEmitter();



exports.setFirstResponder = false;



exports.connect = function(url) {
  return spawnRemoteActor(url)._ref;
};



exports.disconnect = function(ref) {
  var actor;

  if (ref instanceof RemoteRef == false) {
    throw new Error("bad argument. `ref`, expected RemoteRef");
  }

  if ((actor = remotes[ref.id])) {
    actor.destroy();
  }
};



exports.fork = function(filename, params, env) {
  var spawn = require("child_process").spawn;
  var args = [filename];
  var options = { customFds: [0, 1, 2] };
  var child;
  var ref;

  params = params || {};

  if ("url" in params == false) {
    throw new Error("cannot spawn child without connection point");
  }

  for (var param in params) {
    args.push("--" + param + "=" + params[param]);
  }

  if (env) {
    options.env = env;
  }

  child = spawn(process.execPath, args, options); 

  return spawnRemoteActor(params["url"], child)._ref;
};



exports.spawn = function(filename, params) {
  var mod = require(filename);

  if (typeof mod.spawn == "function") {
    return mod.spawn(params);
  } else {
    throw new Error("missing function `spawn`.");
  }
};



exports.locals = function() {
  var keys = Object.keys(locals);
  var index = keys.length;
  var result = [];

  while (index--) {
    result.push(locals[keys[index]]._ref);
  }

  return result;
};



exports.remotes = function() {
  var keys = Object.keys(remotes);
  var index = keys.length;
  var result = [];

  while (index--) {
    result.push(remotes[keys[index]]._ref);
  }

  return result;
};



exports.ping = function(ref, callback) {
  exports.dispatch(ref, "ping", [], callback);
};



exports.exit = function(ref, reason, callback) {
  exports.dispatch(ref, "exit", [reason], callback);
};



exports.kill = function(ref) {
  exports.dispatch(ref, "kill", []);
};



exports.register = function(name, ref) {
  var actor;

  if (typeof name !== "string") {
    throw new Error("bad argument, `name`, expected String");
  }

  if (typeof names[name] !== "undefined") {
    throw new Error("name is already registered");
  }

  if (ref instanceof Ref) {
    if ((actor = locals[ref.id]) || (actor = remotes[ref.id])) {
      if (actor._name) {
        throw new Error("Actor of `ref` already has a name");
      } else {
        actor._name = name;
        actor.monitor && (actor.monitor._name = name);
      }
    } else {
      throw new Error("Ref `ref` is not available");
    }
  } else {
    throw new Error("bad argument, `ref`, expected LocalRef or RemoteRef");
  }

  names[name] = ref;

  if (actor._available) {
    process.nextTick(function() {
      onregister(name, ref);
    });
  }
};



exports.unregister = function(name) {
  var actor;
  var ref;

  if (typeof name !== "string") {
    throw new Error("bad argument, `name`, expected String");
  }

  if ((ref = names[name])) {
    if ((actor = locals[ref.id]) || (actor = remotes[ref.id])) {
      actor._name = null;
      delete names[name];
      if (actor._available) {
        exports.emit("unregister", name, ref);
      }
    }
  }

};



exports.registered = function() {
  return Object.keys(names);
};



exports.whereis = function(name) {
  return names[name] || null;
};



exports.available = function(refOrName) {
  var actor = exports.getActorByRefOrName(refOrName);
  return actor ? actor._available : false;
};



exports.exists = function(refOrName) {
  var actor = exports.getActorByRefOrName(refOrName);
  return actor ? true : false;
};



exports.getActorByRefOrName = function(refOrName) {
  if (refOrName instanceof Ref) {
    return exports.getActorByRef(refOrName);
  } else if (typeof refOrName == "string") {
    return exports.getActorByName(refOrName);
  } else{
    throw new Error("bad argument, `refOrName`, expected Ref or String");
  }
}



exports.getActorByRef = function(ref) {
  if (ref instanceof LocalRef) {
    return locals[ref.id] || null;  
  } else if (ref instanceof RemoteRef) {
    return remotes[ref.id];
  }
  throw new Error("bad argument, `ref`, expected Ref");
};



exports.getActorByName = function(name) {
  var ref = names[name];
  return ref && exports.getActorByRef(ref) || null;
};



exports.registerLocal = function(obj, params) {
  var ref = new LocalRef(uid());

  locals[ref.id] = obj;

  obj._ref = ref;
  obj._name = null;
  obj._available = false;

  if (exports.setFirstResponder) {
    exports.setFirstResponder = false;
    exports.firstResponderRef = ref;
  }

  function announce() {
    obj._available = true;
    onavailable(ref);
    if (obj._name) {
      onregister(obj._name, ref);
    }
  }

  if ((keys = Object.keys(params)).length) {
    process.nextTick(function next(err) {
      if (err) return obj.destroy(err);
      var key = keys.pop();
      key ? obj.set(key, params[key], next) : process.nextTick(announce);
    });
  } else {
    process.nextTick(announce);
  }
  
};



exports.unregisterLocal = function(obj, reason) {
  if (obj._name) {
    exports.unregister(obj._name);
    obj._name = null;
  }

  if (obj._ref) {
    onunavailable(obj._ref, reason || "NORMAL");
    delete locals[obj._ref.id];
    obj._ref = null;
  }

  obj._available = false;
};



exports.registerProtocolHandler = function(protocol, ref) {
  var provider;

  if (typeof protocol !== "string" || !protocol.length) {
    throw new Error("bad argument, `protocol`, expected non-empty string");
  }

  if (ref instanceof LocalRef == false) {
    throw new Error("bad argument, `ref`, expected LocalRef");
  }

  provider = exports.getActorByRef(ref);

  if (!provider || (!provider.connect && !provider.listen)) {
    throw new Error("bad argument, `handler`, must provide connect or listen");
  }

  if (protocol[protocol.length - 1] !== ":") {
    protocol = protocol + ":";
  }

  if (protocolHandlers[protocol]) {
    throw new Error("protocol is already registered `" + protocol + "`");
  }

  protocolHandlers[protocol] = provider;

  exports.emit("protocolAvailable", protocol, ref);
};



exports.unregisterProtocolHandler = function(protocol) {
  var actor;

  if (protocol[protocol.length - 1] !== ":") {
    protocol = protocol + ":";
  }
  
  if (!(actor = protocolHandlers[protocol])) {
    throw new Error("no protocol handler registered for `" + protocol + "´");
  }

  protocolHandlers[protocol] = null;

  exports.emit("protocolAvailable", protocol, actor._ref);
};



exports.getProtocolHandler = function(protocol) {

  if (!protocol) {
    throw new Error("bad argument, `protocol`, expected String");
  }

  if (protocol[protocol.length - 1] !== ":") {
    protocol = protocol + ":";
  }
  
  if (!protocolHandlers[protocol]) {
    throw new Error("no protocol handler registered for `" + protocol + "´");
  }

  return protocolHandlers[protocol]._ref;
};



exports.hasProtocolHandler = function(protocol) {
  return typeof protocolHandlers[protocol] !== "undefined";
};



exports.listen = function(url, callback) {
  var protocol;
  var handler;
  var handlerref;

  if (typeof url !== "string" && typeof url !== "object") {
    throw new Error("bad argument, `url`, expected String or Object");
  }

  url = typeof url == "string" ? require("url").parse(url) : url;

  protocol = url.protocol;

  if (exports.hasProtocolHandler(protocol) == false) {
    if (protocolWaitlist[protocol]) {
      protocolWaitlist[protocol].push(url, callback || function() {});
    } else {
      protocolWaitlist[protocol] = [url, callback || function() {}];
    }
    return;
  }

  handlerref = exports.getProtocolHandler(protocol);
  handler = exports.getActorByRef(handlerref);

  if (typeof handler.listen !== "function") {
    if (callback) {
      callback(new Error("listening is not supported for protocol"));
    }
    return;
  }

  handler.listen(url, exports.dispatch, callback || function() {});
};



exports.unlisten = function(url) {
  var handler;
  var handlerref;
  var protocol;

  if (typeof url !== "string" && typeof url !== "object") {
    throw new Error("bad argument, `url`, expected String or Object");
  }

  url = typeof url == "string" ? require("url").parse(url) : url;

  protocol = url.protocol;

  if (exports.hasProtocolHandler(protocol) == false) {
    return;
  }

  handlerref = exports.getProtocolHandler(protocol);
  handler = exports.getActorByRef(handlerref);

  if (!handler || typeof handler.unlisten !== "function") {
    return;
  }

  handler.unlisten(url);
};



exports.dispatch = function(ref, handler, args, callback) {
  var m;

  if (typeof ref == "string") {
    if (typeof handler !== "string") {
      return dispatchByString(ref, handler, args);
    } else {
      if (typeof args == "function") {
        callback = args;
        args = [];
      }
      if (Array.isArray(args) == false) {
        throw new Error("bad argument, `args`, expected Array");
      }
      return dispatchNamed(ref, handler, args, callback);
    }
  } else {
    if (typeof handler !== "string") {
      throw new Error("bad argument, `handler`, expected string");
    }
    if (typeof args == "function") {
      callback = args;
      args = [];
    }
    if (Array.isArray(args) == false) {
      throw new Error("bad argument, `args`, expected Array");
    }
    if (ref instanceof LocalRef) {
      return dispatchByRef(ref, handler, args, callback);
    } else if (ref instanceof RemoteRef) {
      return dispatchRemote(ref, handler, args, callback);
    } else {
      throw new Error("bad argument, `ref`, expected Ref");
    }
  }
};



exports.isFirstResponder = function(ref) {
  return exports.firstResponderRef === ref;
};



exports.getFirstResponder = function() {
  return exports.firstResponderRef;
};



function Ref(id) {
  this.id = id;
}

exports.Ref = Ref;



function LocalRef(id) {
  Ref.call(this, id);
  Object.freeze(this);
}

exports.LocalRef = LocalRef;
inherits(LocalRef, Ref);

LocalRef.prototype.inspect =
LocalRef.prototype.toString = function() {
  return ["<Actor ", this.id, "@LOCAL>"].join("");
};



function RemoteRef(url) {
  Ref.call(this, url.href);
  this._url = url;
  Object.freeze(this);
}

exports.RemoteRef = RemoteRef;
inherits(RemoteRef, Ref);

RemoteRef.prototype.inspect =
RemoteRef.prototype.toString = function() {
  return ["<RemoteActor ", this.id, ">"].join("");
};



// TODO: Handle closing connections
function Remote(url, child) {
  var self = this;
  var ref = new RemoteRef(url);

  remotes[ref.id] = this;

  this._ref = ref;
  this._name = null;

  this._hash = url.hash && url.hash.substr(1) || null;
  this._available = false;
  this._connection = null;

  function doconnect() {
    var handlerref;
    var handler;

    try {
      handlerref = exports.getProtocolHandler(url.protocol);
    } catch (protocolerr) {
      self._connectTimer = setTimeout(doconnect, 800);
      return;
    }

    handler = exports.getActorByRef(handlerref);

    if (typeof handler.connect !== "function") {
      self._connectTimer = setTimeout(doconnect, 800);
      return;
    }

    handler.connect(url, function(err, conn) {
      var callback;

       if (err) {
         self._connectTimer = setTimeout(doconnect, 800);
         return;
       }

       clearTimeout(self._connectTimer);

       conn.__refcount__ = conn.__refcount__ ? conn.__refcount__ + 1 : 1;
       self._connectTimer = null;
       self._connection = conn;
       self._available = true;
       conn.on("error", function(err) {
         self.destroy(err);
       });
       conn.on("close", function(hadError) {
         if (hadError) {
           self.destroy(new Error("DOWN"));
         }
       });
       onavailable(ref);
    });
  }

  if (child) {
    this._child = child;
    child.once("exit", function (code) {
      self._child = null;
      self.destroy(code && new Error("EXITCODE" + code) || null);
    });
  }

  this._connectTimer = setTimeout(doconnect, 0);
}

exports.Remote = Remote;
inherits(Remote, EventEmitter);



Remote.prototype.dispatch = function(name, args, callback) {
  if (this._hash) {
    name = this._hash + "#" + name;
  }

  return this._connection.dispatch(name, args, callback);
};



Remote.prototype.destroy = function(reason) {
  var ref;
  var didconnect = true;

  if (this._connectTimer) {
    clearTimeout(this._connectTimer);
    this._connectTimer = null;
    didconnect = false;
  }

  if (this._child) {
    this._child.removeAllListeners("exit");
    this._child.kill("SIGKILL");
    this._child = null;
  }

  if (this._connectCallback) {
    this._connectCallback(new Error("destroy"));
    this._connectCallback = null;
  }

  if (this._connection) {
    if (--this._connection.__refcount__ == 0) {
      this._connection.destroy();
    }
    this._connection = null;
  }

  if ((ref = this._ref)) {
    onunavailable(ref, reason || "NORMAL");
    delete remotes[ref.id];
    this._ref = null;
  }

  this._available = false;
};



function spawnRemoteActor(url, child) {
  var remote;

  if (url instanceof RemoteRef) {
    if (remotes[url.id]) {
      return url;
    } else {
      url = url._url;
    }
  } else {
    url = typeof url == "string" ? require("url").parse(url) : url;
  }

  if ((remote = remotes[url.href])) {
    return actor;
  }

  remote = new Remote(url, child);

  return remote;
}



var uid = (function() {
  var pool = new Array(ID_POOL_INITIAL_SIZE);
  var length = ID_POOL_INITIAL_SIZE;

  for (var i = 0; i < ID_POOL_INITIAL_SIZE; i++) {
    pool[i] = i + 1;
  }

  return function(id) {
    if (arguments.length) {
      pool.push(id);
    } else {
      if (pool.length === 0) {
        if (length + 1 == ID_POOL_MAX) {
          throw new Error("MAX_ID reached");
        }
        console.log("Increase uid pool" + length);
        for (var i = length + 1, l = (length * ID_POOL_INCREASE_STEP) + 1;
             i < l; i++) {
          pool.push(i);
        }
        length = l - 1;
      }

      return pool.shift();
    }
  };
})();



function dispatchByString(handler, args, callback) {
  var m;
  var ref;

  if (typeof args == "function") {
    callback = args;
    args = [];
  }

  if (args instanceof Array == false) {
    throw new Error("bad argument, `args expected Array");
  }

  // TODO: Check for URL´s
  if ((m = /^([A-Za-z0-9\_\$]+)\#([A-Za-z0-9\_\$]+)$/.exec(handler))) {
    return dispatchNamed(m[1], m[2], args, callback);
  } else if (Dispatcher.isHandler(Kernel[handler]) == false) {
    if (exports.firstResponderRef) {
      return dispatchByRef(exports.firstResponderRef, handler, args, callback);
    } else {
      return callback && callback(new Error("no handler for " + handler));
    }
  }

  if (callback) {
    args = args.slice(0);
    args[args.length] = callback;
  }

  switch (args.length) {
    case 0: return Kernel[handler]();
    case 1: return Kernel[handler](args[0]);
    case 2: return Kernel[handler](args[0], args[1]);
    case 3: return Kernel[handler](args[0], args[1], args[2]);
    case 4: return Kernel[handler](args[0], args[1], args[2], args[3]);
    case 5: return Kernel[handler](args[0], args[1], args[2], args[3], args[4]);
    default: return Kernel[handler].apply(this, args);
  }
}



function dispatchNamed(name, handler, args, callback) {
  var target;

  if ((target = exports.getActorByName(name))) {
    return target.dispatch(handler, args, callback);
  }

  if (namedinboxes[name]) {
    namedinboxes[name].push([handler, args, callback]);
  } else {
    namedinboxes[name] = [[handler, args, callback]];
  }

  return false;
}



function dispatchByRef(ref, handler, args, callback) {
  var target;

  if ((target = exports.getActorByRef(ref))) {
    return target.dispatch(handler, args, callback);
  }

  if (inboxes[ref.id]) {
    inboxes[ref.id].push([handler, args, callback]);
  } else {
    inboxes[ref.id] = [[handler, args, callback]];
  }
}



function dispatchRemote(ref, handler, args, callback) {
  var target;

  if ((target = exports.getActorByRef(ref)) && target._available) {
    return target.dispatch(handler, args, callback);
  }

  if (inboxes[ref.id]) {
    inboxes[ref.id].push([handler, args, callback]);
  } else {
    inboxes[ref.id] = [[handler, args, callback]];
  }
}



function emptyInbox(messages, target) {
  var message;

  for (var i = 0, l = messages.length; i < l; i++) {
    message = messages[i];
    target.dispatch(message[0], message[1], message[2]);
  }
}



function onprotocolavailable(protocol, ref) {

  if ((list = protocolWaitlist[protocol])) {
    delete protocolWaitlist[protocol];
    for (var i = 0, l = list.length; i<l; i = i + 2) {
      exports.listen(list[0], list[1]);
    }
  }

  exports.emit("protocolAvailable", protocol, ref);
}



function onprotocolunavaialble(protocol, ref) {
  // TODO: Implement me!
  exports.emit("protocolUnavailable", protocol, ref);
}



function onregister(name, ref) {
  var messages;
  if ((messages = namedinboxes[name])) {
    emptyInbox(messages, exports.getActorByRef(ref));
    delete namedinboxes[name];
  }
  exports.emit("register", name, ref);
}



function onavailable(ref) {
  var messages;

  if ((messages = inboxes[ref.id])) {
    emptyInbox(messages, exports.getActorByRef(ref));
    delete inboxes[ref.id];
  }

  exports.emit("available", ref);
}



function onunavailable(ref, reason) {
  exports.emit("unavailable", ref, reason);
}



Kernel = exports.Kernel = { _ref: new LocalRef(uid()) };

Dispatcher.call(Kernel);
Component.call(Kernel);



(function() {
  function cp(s, t) { for (var k in s) { t[k] = s[k]; } }
  cp(Dispatcher.prototype, Kernel);
  cp(Component.prototype, Kernel);
})();



exports.KernelRef = Kernel._ref;
locals[Kernel._ref.id] = Kernel;
exports.register("app", Kernel._ref);

Component.defineParam(Kernel, "url", {
  desc: "Set's or get's the application URL", type: String,
  initial: ""
});



Component.defineParam(Kernel, "title", {
  desc: "Set's or get's title of the application",
  type: String, initial: process.title
});



Component.defineParam(Kernel, "verbose", {
  desc: "Indicates if application should do verbose logging",
  type: Boolean, initial: true
});



Component.defineParam(Kernel, "temp_dir", {
  desc: "Path to directory where app-temporary files should be stored",
  type: String, initial: "/tmp"
});



Dispatcher.defineHandler(Kernel, "kill", {
  notif: function(code) {
    process.nextTick(function() {
      process.exit(code || 0);
    });
  }
});



Dispatcher.defineHandler(Kernel, "exit", {
  sync: function(reason, callback) {
    var self = this.ref;

    exports.locals().forEach(function(ref) {
      if (ref !== self) {
        exports.dispatch(ref, "exit", [reason]);
      }
    });

    exports.remotes().forEach(function(ref) {
      exports.disconnect(ref);
    });

  }
});



var Monitor = require("./monitor").Monitor;

Kernel.map("title", process, "title");
Kernel.map("title", Monitor, "globalName");
Kernel.map("verbose", Monitor.verboseStream, "writable");
Kernel.map("url", function(value, old) {
  if (old && old !== "") {
    exports.unlisten(old);
  }
  if (value && value !== "") {
    exports.listen(value);
  }
});

//
// if (callback) {
//   timeout = opts && opts.timeout || Actor.dispatchTimeout;
//
//   ctx.id = RemoteActor._nextid++;
//   ctx.callback = callback;
//
//   ctx._onTimeout = function() {
//     var index = self._outbox.indexOf(ctx);
//     self._outbox.splice(index, 1);
//     delete self._waithandlers[ctx.id];
//     self._waithandlersLen--;
//     if (self._waithandlersLen == 0 && self._destroying) {
//       self._destroying = false;
//       self.destroy();
//     }
//     try {
//       callback(new Error("timeout"));
//     } catch (err) {
//       self.emit(err);
//     }
//   };
//
//   enroll(ctx, timeout);
// }
