
//
//        Copyright 2010 Johan Dahlberg. All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions
//  are met:
//
//    1. Redistributions of source code must retain the above copyright notice,
//       this list of conditions and the following disclaimer.
//
//    2. Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//
//  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
//  AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
//  THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
//  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//


var FORMAT_RE = /%[sdjem]/g;

var EventEmitter    = require("events").EventEmitter;
var Stream          = require("stream").Stream;
var inherits        = require("util").inherits;


exports.createMonitor = function(name, opts) {

  if (typeof name !== "string") {
    opts = name;
    name = null;
  }

  return new Monitor(name, opts || {});
};


function MonitorStream() {
  Stream.call(this);

  this.readable = true;
  this.writable = true;

  this._buffer = null;
}


exports.MonitorStream = MonitorStream;
inherits(MonitorStream, Stream);


MonitorStream.prototype.write = function(d) {
  var data;

  if (typeof d == "undefined" || this.writable == false) {
    return false;
  }

  if (Buffer.isBuffer(d)) {
    data = d;
  } else if (typeof d == "string") {
    data = new Buffer(d, "utf8");
  }

  return this._writeOut(data);
};


MonitorStream.prototype._writeOut = function(data) {
  var buffer;
  if ((buffer = this._buffer)) {
    buffer.push(data);
    return false;
  } else {
    if (this.ondata) this.ondata(data, 0, data.length);
    if (this._events && this._events["data"]) {
      this.emit("data", data);
    }
  }
};


MonitorStream.prototype.pause = function() {
  if (!this._buffer) {
    this._buffer = [];
  }
};


MonitorStream.prototype.resume = function() {
  var buffer = this._buffer;
  var index;
  if (buffer) {
    this._buffer = null;
    for (var i = 0, l = buffer.length; i < l; i++) {
      this._writeOut(buffer[i]);
    }
  }
};


MonitorStream.prototype.end = function() {
  this.emit("end");
};


MonitorStream.prototype.destroy = function(err) {
  if (err) {
    this.emit(err);
  }
  this.emit("close");
};


function Monitor(name) {
  this._name = name;
  this._logStream = new MonitorStream();
  this._verboseStream = new MonitorStream();
  this._errorStream = new MonitorStream();
}

exports.Monitor = Monitor;

Monitor.globalName = null;
Monitor.silent = false;
Monitor.showErrorStacks = true;

Monitor.logStream = new MonitorStream();
Monitor.logStream.pipe(process.stdout);

Monitor.verboseStream = new MonitorStream();
Monitor.verboseStream.pipe(process.stdout);

Monitor.errorStream = new MonitorStream();
Monitor.errorStream.pipe(process.stderr);

Monitor.defineStream = function(target, name, descriptor) {
  var streamname;

  if (typeof target == "undefined" || target == null) {
    throw new Error("bad argument, `target`, expected Object");
  }

  if (typeof name !== "string") {
    throw new Error("bad argument, `name`, expected String");
  }

  streamname = "_" + name;

  target[name] = function() {
    var frm;
    if (this.silent) { return; }
    if (!this[streamname]) {
      this[streamname] = new MonitorStream();
      if (descriptor.pipe) {
        this[streamname].pipe(descriptor.pipe);
      }
    }
    frm = this._format.apply(this, arguments) + "\n";
    this[streamname].write(frm);
  }
};

Object.defineProperty(Monitor.prototype, "name", {
  enumerable: true,
  get: function() {
    if (Monitor.globalName && this._name) {
      return Monitor.globalName + "(" + this._name + ")";
    }
    return this._name || Monitor.globalName;
  },
  set: function(value) {
    this._name = value;
  }
});

Object.defineProperty(Monitor.prototype, "silent", {
  enumerable: true,
  get: function() { return Monitor.silent; }
});

Object.defineProperty(Monitor.prototype, "showErrorStacks", {
  enumerable: true,
  get: function() { return Monitor.showErrorStacks; }
});

Monitor.defineStream(Monitor.prototype, "log", { pipe: Monitor.logStream });
Monitor.defineStream(Monitor.prototype, "error", { pipe: Monitor.errorStream });
Monitor.defineStream(Monitor.prototype, "verbose", { pipe: Monitor.verboseStream });

Monitor.prototype._format = function(f) {
  var self = this;
  var util = require('util');
  var i = 1;
  var args;
  var str;
  var name;

  if (typeof f !== 'string') {
    var objects = [];
    for (var i = 0; i < arguments.length; i++) {
      objects.push(util.inspect(arguments[i]));
    }
    return objects.join(' ');
  }


  if ((name = this.name)) {
    f = "%s: " + f;
    args = Array.prototype.slice.call(arguments, 1);
    args = [f, name].concat(args);
  } else {
    args = arguments;
  }

  str = String(f).replace(FORMAT_RE, function(x) {
    switch (x) {
      case '%s': return String(args[i++]);
      case '%d': return Number(args[i++]);
      case '%j': return JSON.stringify(args[i++]);
      case '%e': return self._formatError(args[i++]);
      default:
        return x;
    }
  });
  for (var len = args.length, x = args[i]; i < len; x = args[++i]) {
    if (x === null || typeof x !== 'object') {
      str += ' ' + x;
    } else {
      str += ' ' + util.inspect(x);
    }
  }
  return str;
};

Monitor.prototype._formatError = function(e) {
  var stack;
 if (e instanceof Error) {
   if (e.stack && this.showErrorStacks) {
     stack = e.stack.split("\n");
     return stack.join("\n");
   } else {
     return new String(e.message || "unknown error");
   }
 } else {
   return new String(e);
 }
};