
//
//        Copyright 2010 Johan Dahlberg. All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions
//  are met:
//
//    1. Redistributions of source code must retain the above copyright notice,
//       this list of conditions and the following disclaimer.
//
//    2. Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//
//  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
//  AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
//  THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
//  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

var inherits              = require("util").inherits;

var kernel                = require("./kernel");
var Kernel                = kernel.Kernel;
var LocalRef              = kernel.LocalRef;
var getActorByRefOrName   = kernel.getActorByRefOrName;

var component             = require("./component");
var Dispatcher            = require("./dispatcher").Dispatcher;


var PARAM_QUERY_RE = /^([\w\_\$]+)(?:\.([\w\_\$]+))|^([\w\_\$]+)/;
var FILTER_QUERY_RE = /^(\*|[\w\_\$]+)(?:\.(\*|[\w\_\$]+))|^(\*|[\w\_\$]+)/;

var listeners = {};
var waitinglist = {};


exports.map = function(ref, key, dest, prop) {
  var target;
  var callback;
  var tmp;
  var value;

  if (typeof ref == "string") {
    if (ref.indexOf(".") != -1) {
      prop = dest;
      dest = key;
      tmp = ref.split(".");
      ref = tmp[0];
      key = tmp[1];
    } else {
      key = ref;
      ref = kernel.KernelRef.id.toString();
    }
  } else if (ref instanceof LocalRef) {
    ref = ref.id.toString();
  } else {
    throw new Error("bad argument, `ref`, expected ref");
  }

  if (!dest) {
    throw new Error("bad argument, `dest`, expected Object or Function");
  }

  if (typeof prop == "string" || typeof prop == "function") {
    callback = attachObjectListener(ref, key, dest, prop);
  } else if (typeof dest == "function"){
    callback = attachListener(ref, key, dest);
  } else {
    throw new Error("bad arguments");
  }

  target = kernel.getActorByRefOrName(ref);

  if (target) {
    if ((value = target.__get__(key)) !== void(0)) {
      callback(value, value);
    }
  } else {
    if (!waitinglist[ref]) {
      waitinglist[ref] = [callback];
    } else {
      waitinglist[ref].push(callback);
    }
  }
};



exports.unmap = function() {
  // TODO: implement this!
};



exports.getValidator = function(descriptor) {

  if (typeof descriptor.validator == "function") {
    return customValidator(descriptor);
  }

  switch (descriptor.type) {
    case Number: return numberValidator(descriptor);
    case Boolean: return booleanValidator(descriptor);
    case String: return stringValidator(descriptor);
    default: return basicValidator(descriptor);
  }

  throw new Error("unknown field type");
};


kernel.on("register", function(name, ref) {
  var list;

  if ((list = waitinglist[name])) {
    delete waitinglist[name];
    emptyWaitinglist(list, kernel.getActorByRef(ref));
  }
});



kernel.on("unregister", function(name, ref) {
  var list;

  if ((list = listeners[name])) {
    readdToWaitlist(name, list);
  }
});



kernel.on("available", function(ref) {
  var list;

  if ((list = waitinglist[ref.id])) {
    delete waitinglist[ref.id];
    emptyWaitinglist(list, kernel.getActorByRef(ref));
  }
});



kernel.on("unavailable", function(ref) {
  var list;

  if ((list = listeners[ref.id])) {
    readdToWaitlist(name, list);
  }

});



Dispatcher.defineHandler(Kernel, "get", {
  async: function(param) {
    var callback = Dispatcher.findCallback(arguments);
    var target;
    var ref;
    var m;

    if (typeof param !== "string") {
      return callback && callback(new Error("bad argument, `param`, expected String"));
    }

    if ((m = PARAM_QUERY_RE.exec(param))) {
      ref = m[1] || kernel.KernelRef;
      param = m[2] || m[3];
    } else {
      return callback(new Error("bad param query `" + param + "`"));
    }

    if (!(target = kernel.getActorByRef(ref))) {
      return callback(new Error("invalid target ref"));
    }
  
    if (target.hasParam(param) == false) {
      return callback(new Error("param `" + param + "` was not found"));
    }

    return callback(null, target.__get__(param));
  } 
});



Dispatcher.defineHandler(Kernel, "set", {
  async: function(param, value) {
    var callback = Dispatcher.findCallback(arguments);
    var target;
    var ref;
    var m;

    if (typeof value == "function") {
      return callback && callback(new Error("bad argument, `value`"));
    }

    if (typeof param !== "string") {
      return callback && callback(new Error("bad argument, `param`, expected String"));
    }

    if ((m = PARAM_QUERY_RE.exec(param))) {
      ref = m[1] || kernel.KernelRef;
      param = m[2] || m[3];
    } else {
      return callback(new Error("bad param query `" + param + "`"));
    }

    if (!(target = kernel.getActorByRef(ref))) {
      return callback(new Error("invalid target ref"));
    }
  
    if (target.hasParam(param) == false) {
      return callback(new Error("param `" + param + "` was not found"));
    }

    target.__validate__(param, value, function(err) {
      if (err) return callback(err);
      target.__set__(param, value);
      callback(null, value);
    });
  } 
});



Dispatcher.defineHandler(Kernel, "filter", {
  async: function(filter) {
    var callback = Dispatcher.findCallback(arguments);
    var locals = kernel.locals();
    var index = locals.length;
    var tasks = [];
    var lefthand;
    var righthand;
    var m;

    if (typeof filter == "string" && (m = FILTER_QUERY_RE.exec(filter))) {
      lefthand = m[1] || m[3];
      lefthand = lefthand == "*" ? null : new RegExp("^" + (lefthand));
      righthand = m[2];
    }

    while (index--) {
      tasks.push((function(target) {
        if (!lefthand || (lefthand.test(target._name) ||
                         (lefthand.test(target._ref.id)))) {
          return function(result, next) {
            var params = target.__filter__(righthand);
            var name = target._name || target._ref.id;
            for (var key in params) {
              result[name + "." + key] = params[key];
            }
            next(result)
          }
        } else {
          return function (result, next) { next(result); }
        }
      })(kernel.getActorByRef(locals[index])));
    }

    (function next(result) {
      var task = tasks.shift();
      return task ? task(result, next) : callback && callback(null, result);
    })({});
  }
});



Dispatcher.defineHandler(Kernel, "params", {
  sync: function(params) {
    var callback = Dispatcher.findCallback(arguments);
    var tasks = [];
    var index;
    var target;
    var ref;
    var m;

    if (typeof params !== "object") {
      return callback("bad argument, `params`, expected Object");
    }

    for (var param in params) {
      if ((m = PARAM_QUERY_RE.exec(param))) {
        ref = m[1] || kernel.KernelRef;
        target = getActorByRefOrName(ref);
        if (!target) {
          return callback(new Error("actor not available %s", ref));
        }
        if (target.hasParam(m[2] || m[3]) == false) {
          return callback(new Error("bad param, `" + param +"`, invalid"));
        }
        tasks.push((function(target, param, key) {
          function task(next) {
            target.__validate__(param, params[key],
              function(err, value) {
                if (err) {
                  err = new Error(key + ": " + err.message);
                  return callback(err);
                }
                task.value = value;
                next();
            });
          };
          task.key = key;
          task.target = target;
          task.param = param;
          return task;
        })(target, m[2] || m[3], param));
      } else {
        return callback(new Error("bad param query `" + param + "`"));
      }
    }

    index = tasks.length;

    (function next() {
      var task = tasks[--index];
      var result;
      if (task) task(next);
      else {
        index = tasks.length;
        result = {};
        while (index--) {
          task = tasks[index];
          result[task.key] = task.value;
          task.target.__set__(task.param, task.value);
        }
        callback(null, result);
      }
    })();
    
  }
});



function attachListener (ref, key, callback) {
  var emitkey = "param." + ref + "." + key;
  var objects;
  var props;

  if (!(objects = listeners[ref])) {
    listeners[ref] = objects = {};
  }

  if (!(props = objects[key])) {
    objects[key] = callback;
    component.on(emitkey, function(value, oldvalue) {
      var props = objects[key];
      var index;
      if (typeof props == "function") {
        props(value, oldvalue);
      } else {
        index = props.length;
        while (index--) {
          props[index](value, oldvalue);
        }
      }
    });
  } else if (Array.isArray(props)) {
    props.push(callback);
  } else {
    objects[key] = [props, callback];
  }

  callback._key = key;

  return callback;
}



function attachObjectListener (ref, key, obj, prop) {
  var listener;

  if (typeof prop == "function") {
    listener = function(value, old) { prop.call(obj, value, old); }
  } else {
    listener = function(value) { obj[prop] = value; }
  }

  listener._obj = obj;
  listener._prop = prop;

  return attachListener(ref, key, listener);
}



function emptyWaitinglist(list, target) {
  var index = list.length;
  var callback;
  var value;

  while (index--) {
    callback = list[index];
    if ((value = target.__get__(callback._key)) !== void(0)) {
      callback(value, value);
    }
  }
}



function readdToWaitlist(ref, list) {
  var newlist;
  var keys;
  var index;
  var props;
  var callbacks;

  newlist = [];
  props = listeners[ref];
  keys = Object.keys(props);
  index = keys.length;
  while (index--) {
    callbacks = props[keys[index]];
    if (typeof callbacks == "function") {
      newlist.push(callbacks);
    } else {
      newlist = newlist.concat(callbacks);
    }
  }
  waitinglist[ref] = newlist;
}



function basicValidator(opts) {
  return function(value, callback) {
    var err;

    if (opts.readonly == true) {
      err = new Error("param is readonly");
      return callback(err);
    }

    callback(null, value);
  }
}



function customValidator(opts) {
  var validator = opts.validator;
  return function(value, callback) {
    var err;

    if (opts.readonly == true) {
      err = new Error("param is readonly");
      return callback(err);
    }

    return validator.call(this, value, callback);
  }
}



function numberValidator(opts) {
  var max = opts.max || Number.MAX_VALUE;
  var min = opts.min || Number.MIN_VALUE;
  return function (value, callback) {
    var err;

    if (opts.readonly == true) {
      err = new Error("param is readonly");
      return callback(err);
    }

    if (typeof value !== "number") {
      value = parseInt(value);
    }

    if (isNaN(value)) {
      err = new Error("bad argument, expected Number");
      return callback(err);
    }

    if (value > max || value < min) {
      err = new Error("invalid range, expected [" + min + "-" + max + "]");
      return callback(err);
    }

    return callback(null, value);
  };
}



function stringValidator(opts) {
  var allowEmpty = opts.empty || true;
  return function (value, callback) {
    var err;

    if (opts.readonly == true) {
      err = new Error("param is readonly");
      return callback(err);
    }

    if (typeof value !== "string") {
      value = (new String(value)).toString();
    }

    if (value.length == 0 && !allowEmpty) {
      return callback(new Error("value cannot be `empty`"));
    }

    return callback(null, value);
  };
}



function booleanValidator (opts) {
  return function (value, callback) {
    var err;

    if (opts.readonly == true) {
      err = new Error("param is readonly");
      return callback(err);
    }

    if (typeof value !== "boolean") {
      switch (value) {
        case 0: value = false; break;
        case 1: value = true; break;
        case "yes": value = true; break;
        case "no": value = false; break;
        case "true": value = true; break;
        case "false": value = false; break;
        default: err = new Error("invalid value, expected true or false");
      }
    }

    return callback(err, value);
  };
}
