
//
//        Copyright 2010 Johan Dahlberg. All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions
//  are met:
//
//    1. Redistributions of source code must retain the above copyright notice,
//       this list of conditions and the following disclaimer.
//
//    2. Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//
//  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
//  AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
//  THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
//  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

var inherits              = require("util").inherits;

var Actor                 = require("./actor").Actor;

var register              = require("./kernel").register;

var Stream                = require("net").Stream;
var Server                = require("net").Server;
var inherits              = require("util").inherits;


exports.spawn = function(params) {
  var instance = new RpcActor(params);
  register("rpc", instance.ref);
  return instance.ref;
};



function RpcActor() {
  Actor.call(this);

  require("./kernel").registerProtocolHandler("psock:", this.ref);
  require("./kernel").registerProtocolHandler("ptcp:", this.ref);

  this._servers = {};
  this._clients = {};
}

exports.RpcActor = RpcActor;
inherits(RpcActor, Actor);



Actor.defineHandler(RpcActor.prototype, "exit", {
  sync: function(reason, callback) {
    var keys = Object.keys(this._servers);
    var index = keys.length;
    var key;

    while (index--) {
      key = keys[index];
      this._servers[key].destroy();
    }

    keys = Object.keys(this._clients);
    index = keys.length;

    while (index--) {
      key = keys[index];
      this._clients[key].destroy();
    }

    this._clients = {};
    this._servers = {};

    this.destroy();
  }
});



Actor.defineHandler(RpcActor.prototype, "listen", {
  sync: function(url, handler, callback) {
    var connid = url.protocol + (url.host || url.pathname);
    var self = this;
    var servers = this._servers;
    var server;
    var host;
    var port;

    if (servers[connid]) {
      return;
    }

    server = new RpcServer();
    server._listeningCallbacks = [callback];
    servers[connid] = server;

    if (url.protocol == "ptcp:") {
      host = url.hostname.length ? url.hostname : undefined;
      port = url.port || this._defaultTcpPort;
    } else if (url.protocol == "psock:") {
      port = url.pathname;
    }

    server.on("listening", function() {
      self.verb("listening for connections on %s", url.href);
      var callbacks = server._listeningCallbacks;
      server._listeningCallbacks = null;
      server.messageHandler = handler;
      callbacks && callbacks.forEach(function(callback) {
        callback(null, server);
      });
    });

    server.on("error", function(err) {
      var msg = err.message || err.toString();
      self.err("error on `%s`, error: `%s`", url.href, msg);
      var callbacks = server._listeningCallbacks;
      server._listeningCallbacks = null;
      callbacks && callbacks.forEach(function(callback) {
        callback(err);
      });
    });

    server.on("clientConnect", function(sock) {
      self.verb("client `%s` connected", sock.url);
    });

    server.on("clientError", function(sock, err) {
      var msg = err.message || "unknown error";
      var url = sock.url;
      self.err("client `%s` raised an error \"%s\"", url, msg);
    });

    server.on("clientDisconnect", function(sock) {
      self.verb("client `%s` disconnected", sock.url);
    });

    server.listen(port, host);
  }
});



Actor.defineHandler(RpcActor.prototype, "unlisten", {
  sync: function(url, callback) {
    var connid = url.protocol + (url.host || url.pathname);
    var servers = this._servers;
    var server;

    if (server = servers[connid]) {
      delete servers[connid];
      server.destroy();
    }

    callback && callback();
  }
});



Actor.defineHandler(RpcActor.prototype, "connect", {
  sync: function(url, callback) {
    var connid = url.protocol + (url.host || url.pathname);
    var self = this;
    var clients = this._clients;
    var client;
    var host;
    var port;

    if ((client = this._clients[connid])) {
      if (client._connectCallbacks) {
        client._connectCallbacks.push(callback);
        return;
      }
      return callback(null, client);
    }

    client = new RpcClient();
    client._connectCallbacks = [callback];
    clients[url.href] = client;

    if (url.protocol == "ptcp:") {
      host = url.hostname.length ? url.hostname : undefined;
      port = url.port || this._defaultTcpPort;
    } else if (url.protocol == "psock:") {
      port = url.pathname;
    }

    client.on("connect", function() {
      var callbacks = client._connectCallbacks;
      client._connectCallbacks = null;
      callbacks.forEach(function(callback) {
        callback(null, client);
      });
    });

    client.on("error", function(err) {
      var callbacks = client._connectCallbacks;
      client._connectCallbacks = null;
      callbacks && callbacks.forEach(function(callback) {
        callback(err);
      });
    });

    client.on("close", function() {
      var callbacks = client._connectCallbacks;
      client._connectCallbacks = null;
      callbacks && callbacks.forEach(function(callback) {
        callback(new Error("disconnected"));
      });
      delete clients[connid];
    });


    client.connect(port, host);
  }
});



function RpcServer() {
  Server.call(this);

  this._messageHandlers = [];
  this._clients = [];

  this.on("connection", this.__conncetion__);
}

inherits(RpcServer, Server);



RpcServer.prototype.__conncetion__ = function(sock) {
  var self = this;
  var clients = this._clients;
  var url;

  if (sock.type == "unix") {
    sock.url = "psock://" + this.path;
  } else {
    sock.url = "ptcp://" + sock.remoteAddress + ":" + sock.remotePort;
  }

  clients.push(sock);

  messageParserImpl(sock);

  sock.onmessage = function(cmd, args, ack) {
    if (ack) {
      self.messageHandler(cmd, args, function() {
        var buff;
        var resp = Array.prototype.slice.call(arguments);

        buff = encodeMessage("", resp, ack);

        try {
          sock.write(buff);
        } catch (ioerr) {
          sock.destroy(ioerr);
        }
      });
    } else {
      self.messageHandler(cmd, args);
    }
  };
  
  sock.on("error", function(err) {
    self.emit("clientError", this, err);
  });

  sock.on("close", function() {
    var index = clients ? clients.indexOf(this) : -1;

    if (index !== -1) {
      clients.splice(index, 1);
    }

    this.ondata = null;
    this.onmessage = null;
    
    self.emit("clientDisconnect", this);
  });

  this.emit("clientConnect", sock);
}



RpcServer.prototype.addMessageHandler = function(handler) {
  this._messageHandlers.push(handler);
};



RpcServer.prototype.removeMessageHandler = function(handler) {
  var index = this._messageHandlers.indexOf(handler);

  if (index !== -1) {
    this._messageHandlers.splice(index, 1);
  }

  if (this._messageHandlers.length == 0) {
    this.destroy();
  }
};



RpcServer.prototype.destroy = function() {
  var clients = this._clients;
  var len = clients && clients.length;

  if (len) {
    clients.forEach(function(sock) {
      sock.destroy();
    });
  }
  this.close();
};



function encodeMessage(cmd, args, ack) {
  var msg;
  var buffer;
  var graph;
  var flags;
  var index;
  var arg;
  var ctor;
  var fd;

  flags = new Array(args.length);

  index = args.length;

  while (index--) {
    arg = args[index];

    ctor = (arg === null || arg === undefined) ? arg : arg.constructor;

    switch (ctor) {

      case Error:
        flags[index] = "E";
        args[index] = arg.message || arg.toString();
        break;

      case Stream:
        flags[index] = "F";
        fd = args[index] = arg.fd;
        break;

      default:
        flags[index] = " ";
        break;
    }
  }

  msg = [flags.join(""), cmd, args, ack];

  graph = JSON.stringify(msg);
  buffer = new Buffer(graph + "\n");
  buffer.fd = fd;

  return buffer;
}



function decodeMessage(data, start, end) {
  var msg;
  var buffer;
  var graph;
  var args;
  var flags;
  var flag;

  msg = JSON.parse(data.slice(start, end));

  if (Array.isArray(msg) == false) {
    throw new Error("expected array");
  }

  flags = msg[0];
  args = msg[2];
  index = args.length;

  while (index--) {
    flag = flags[index];

    if (flag === " ") continue;

    switch (flag) {

      case "E":
        args[index] = new Error(args[index]);
        break;

      case "F":
        if (msg.fdindex) {
          throw new Error("multiple FDs");
        }
        msg.fdindex = index + 1;
        break;
    }
  }

  return msg;
}



function RpcClient(opts) {
  Stream.call(this, opts);

  this._messageHandlers = [];

  this._nextack = 1;
  this._waithandlers = {};
  this._waithandlersLen = 0;
  this._waithandlers = {};

  messageParserImpl(this);
}

exports.RpcClient = RpcClient;
inherits(RpcClient, Stream);



RpcClient.prototype.onmessage = function(cmd, args, ack) {
  var cb;

  if ((cb = this._waithandlers[ack])) {
    this._waithandlersLen--;
    delete this._waithandlers[ack];
    if (this._waithandlersLen == 0 && this._destroying) {
      this._destroying = false;
      this.destroy();
    }
    try {
      switch (args.length) {
        case 0: return cb(); break;
        case 1: return cb(args[0]); break;
        case 2: return cb(args[0], args[1]); break;
        case 3: return cb(args[0], args[1], args[2]); break;
        case 4: return cb(args[0], args[1], args[2], args[3]); break;
        case 5: return cb(args[0], args[1], args[2], args[3], args[4]); break;
        default: cb.apply(null, args); break;
      }
    } catch (err) {
      console.log(err);
      this.emit("error", err);
    }
    return true;
  }

  return false;
};



RpcClient.prototype.dispatch = function(name, args, callback) {
  var ack;
  var buff;
  var flushed;

  if (callback) {
    ack = this._nextack++;
  }

  buff = encodeMessage(name, args, ack);

  try {
    flushed = this.write(buff, buff.fd);
    if (ack) {
      this._waithandlers[ack] = callback;
      this._waithandlersLen++;
    }
  } catch (ioerror) {
    console.log(ioerror);
    callback && callback(ioerror);
    this.destroy(ioerror);
    return false;
  }
  
};



function messageParserImpl(sock) {
  var buffer = "";
  var fdqueue = [];
  var fdwaitqueue = [];

  if (sock.type == "unix") {
    sock.on("fd", function(fd) {
      var msg;

      if (fdwaitqueue.length) {
        msg = fdwaitqueue.shift();
        msg[2][msg.fdindex] = fd;
        sock.onmessage(msg[1], msg[2], msg[3]);
      } else {
        fdqueue.push(fd);
      }
    });
  }

  function waitforfd(msg) {
    if (fdqueue.length) {
      msg[2][msg.fdindex] = fdqueue.shift();
      sock.onmessage(msg[1], msg[2], msg[3]);
    } else {
      fdwaitqueue.push(msg);
    }
  }

  sock.ondata = function(data, start, end) {
    var pos = 0;
    var index;
    var msg;
    
    try {

      if (buffer) {
        buffer = buffer + data.toString("utf8", start, end);
      } else {
        buffer = data.toString("utf8", start, end);
      }

      while ((index = buffer.indexOf("\n", pos)) !== -1) {

        msg = decodeMessage(buffer, pos, index);

        // Special case, we are expecting a FD. Pause the read-loop and 
        // wait for a FD to arraive
        if (msg.fdindex) {

          msg.fdindex--;
          if (sock.type !== "unix") {
            sock.destroy(new Error("received fd from non-unix sock"));
            return;
          }

          waitforfd(msg);

        } else {
          sock.onmessage(msg[1], msg[2], msg[3]);
        }

        pos = index + 1;
      }

      if (pos < buffer.length) {
        buffer = buffer.substr(pos);
      } else {
        buffer = null;
      }

    } catch (err) {
      console.log(err.stack);
      sock.destroy(err);
      return;
    }
  };
}